/*
PRICE was written by Graham Ruby.
This file is part of PRICE.

PRICE is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

PRICE is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with PRICE.  If not, see <http://www.gnu.org/licenses/>.
*/

/* This is the programmatic interface.
 */


#ifndef ASSEMBLER_H
#define ASSEMBLER_H

#include <set>
#include <algorithm>
#include <vector>
#include "ScoredSeq.h"
#include "ReadFile.h"
#include "ParamsMapping.h"
#include "ParamsCycleManagement.h"
#include "ParamsAlignment.h"
#include "AssemblyJob.h"
#include "AssemblyJobFactory.h"
#include "ExtendCycle.h"
#include "OutputFile.h"
#include "ParameterizedReadFile.h"
#include "ParameterizedInitialFile.h"
#include "EcoFilter.h"
#include "AssemblerListener.h"
#include "ParamsDeBruijn.h"
#include "ReadPairFilter.h"
using namespace::std;

class Assembler {

 public:
  Assembler();
  Assembler(ParamsCycleManagement* cycleInfo);
  Assembler(int totalCycles);
  Assembler(int totalCycles, long maxContigLinkages);
  ~Assembler();

  void runAssembly();
  void getContigs(set<ScoredSeq*>* outputContigs);

  long getTotalSize();
  long getMaxSize();
  long getN50();

  static ParamsMinOverlap* getDefaultParamsMinOverlap();

  void addInitialFile(string filename, int numSteps, int cyclesPerStep, float countFactor);
  void addInitialFile(long numSeqsToUse, string filename, int numSteps, int cyclesPerStep, float countFactor);
  void addInitialFileNoTarget(string filename, int numSteps, int cyclesPerStep, float countFactor);
  void addInitialFileNoTarget(long numSeqsToUse, string filename, int numSteps, int cyclesPerStep, float countFactor);

  // single files, alternating pairs
  void addReadFile(string filename, long ampSize);
  void addReadFile(string filename, long ampSize, float fractId);
  void addReadFileLimitUse(int startCycle, int numCycles, string filename, long ampSize);
  void addReadFileLimitUse(int startCycle, int numCycles, string filename, long ampSize, float fractId);
  // two files
  void addReadFile(string filenameA, string filenameB, long ampSize);
  void addReadFile(string filenameA, string filenameB, long ampSize, float fractId);
  void addReadFileLimitUse(int startCycle, int numCycles, string filenameA, string filenameB, long ampSize);
  void addReadFileLimitUse(int startCycle, int numCycles, string filenameA, string filenameB, long ampSize, float fractId);

  void addOutputFile(string filename);

  void addBadSequenceFilter(string badSeqFile, float fractId);
  void addHomopolymerFilter(long maxLength);

  void addLengthFilter(long minLength, int cyclesToSkip);
  void addTargetMode(float minFractId, int cyclesToSkip, bool fullFile = false);
  void addTargetMode(float minFractId, int cyclesToSkip,
		     int numFilterCycles, int numSkipCycles, bool fullFile = false);

  void verboseLogFile(string filename);
  void makeLogNull();
  void makeLogVerbose();

  void setMaxThreadsPerFile(int maxThreadsPerFile);
  void setParamsDeBruijn(ParamsDeBruijn* paramsDeBruijn);
  void setMiniAlignParams(ParamsAlignment* params);
  void setMetaAlignParams(ParamsAlignment* params);

 private:
  void constructorHelper();
  // consolidates all of the various inputs and checks for inconsistencies
  void runSetup();
  // cleans up at the end of a run; different from the destructor, it will leave params intact
  void runCleanup();


  // output files as they are being collected
  set<OutputFile*> _outfileSet;
  // the one that will be used; should be generated when assembly is run and deleted at the end of the run
  OutputFile* _outfile;

  // these two are used for set-up
  AssemblerListener* _stdOutListener;
  AssemblerListener* _fileListener;
  // this one is used by the job
  AssemblerListener* _listener;


  // If numInitialContigs==0, then all of the contigs in all the files will be used.
  // Same goes for if numInitialContigs > the num of sequences in the files.
  void generateInitialContigs(int cycleNum);
  void extendOneCycle(int cycleNum);
  void writeOutfile(string addendum);
  void filterOutputContigs(int cycleNum);
  bool reverseN50Helper(long i,long j);
  // helps sort sequences - only modifies leftover by adding components of current
  void findLeftoverContigs(set<ScoredSeq*>* current, set<ScoredSeq*>* prior, set<ScoredSeq*>* leftover);

  // temp storage
  set<ParameterizedReadFile*> _inputPairedReadFiles;
  set<ParameterizedInitialFile*> _inputInitialContigFiles;
  // will not be targeted
  set<ParameterizedInitialFile*> _inputInitialContigFilesNoTarget;
  // possibly modified, used for actual run
  set<ParameterizedReadFile*> _pairedReadFiles;
  set<ParameterizedInitialFile*> _initialContigFiles;

  // collects the input
  vector<ReadPairFilter*> _waitingRpf;
  // actually used in the job
  ReadPairFilter* _rpf;

  // so that contigs can be added at the last minute
  EcoFilterInitialContigMatch* _waitingTargetMode;
  // is created immediately; EcoFilters are added to it directly
  ParamsCycleManagement* _cycleInfo;

  set<ScoredSeq*> _priorCycleContigs;
  set<ScoredSeq*> _currentCycleContigs;
  set<ScoredSeq*> _priorCycleLeftovers;
  set<ScoredSeq*> _currentCycleLeftovers;

  ParamsAlignment* _miniAlignParams;
  ParamsAlignment* _metaAlignParams;
  ParamsDeBruijn* _paramsDeBruijn;



  int _maxThreadsPerFile;
};

#endif
